#lang scheme
(require "test.ss")

;; A board is a vector of cell.
;;  The cells are in order: left-to-right then top-to-bottom
;; A cell is
;;  - number
;;  - #f
;; A dimen is an integer for a board dimension (m or n)
;; A pos is an integer for a board cell position,
;   0 inclusive to (* m n m n) exclsuive

;; rows : board dimen dimen -> (listof (listof cell))
(define (rows board m n)
  (let loop ([l (vector->list board)])
    (if (null? l)
        null
        (cons (take l (* m n))
              (loop (list-tail l (* m n)))))))

;; cols : board dimen dimen -> (listof (listof cell))
(define (cols board m n)
  (apply map list (rows board m n)))

;; sections : board dimen dimen -> (listof (listof cell))
(define (sections board m n)
  (let ([w (* m n)])
    (for*/list ([j (in-range m)]
                [i (in-range n)])
      (for*/list ([l (in-range n)]
                  [k (in-range m)])
        (vector-ref board (+ (* j w n)
                             (* i m)
                             k
                             (* l w)))))))

;; rows : board pos cell -> board
(define (update-cell board pos v)
  (let ([board2 (list->vector (vector->list board))])
    (vector-set! board2 pos v)
    board2))

; ----------------------------------------

;; ok? : board dimen dimen -> boolean
(define (ok? board m n)
  (andmap ok-set?
          (append
           (rows board m n)
           (cols board m n)
           (sections board m n))))

;; ok-set? : (listof cell) -> boolean
;;  Assumes that the cell values are no more than the
;;  list length.
(define (ok-set? l)
  (let* ([len (length l)]
         [v (make-vector (add1 len) #f)])
    (andmap (lambda (e)
              (or (not e)
                  (and (not (vector-ref v e))
                       (vector-set! v e #t)
                       #t)))
            l)))

;; ----------------------------------------

;; solve : board dimen dimen pos -> board-or-false
(define (solve board m n pos)
  (cond
    [(not (ok? board m n)) #f]
    [(= pos (* m n m n)) board]
    [(vector-ref board pos)
     (solve board m n (add1 pos))]
    [else
     (let loop ([i 0])
       (cond
         [(= i (* m n)) #f]
         [else
          (or (solve (update-cell board pos (add1 i))
                     m n (add1 pos))
              (loop (add1 i)))]))]))

;; ----------------------------------------

;; solve/k : board dimen dimen pos (board (-> X) -> X) (-> X) -> X
(define (solve/k board m n pos success-k fail-k)
  (cond
    [(not (ok? board m n)) (fail-k)]
    [(= pos (* m n m n)) (success-k board fail-k)]
    [(vector-ref board pos)
     (solve/k board m n (add1 pos) success-k fail-k)]
    [else
     (let loop ([i 0])
       (cond
         [(= i (* m n)) (fail-k)]
         [else
          (solve/k (update-cell board pos (add1 i))
                   m n (add1 pos)
                   success-k
                   (lambda ()
                     (loop (add1 i))))]))]))

(define (new-solve board m n pos)
  (solve/k board m n  0
           (lambda (board fail-k) board)
           (lambda () #f)))

;; ----------------------------------------
;; Tests

(define ex0 #(a b c  d e f
              g h i  j k l
        
              m n o  p q r
              s t u  v w x
              
              y z 0  1 2 3
              4 5 6  7 8 9))


(test (rows ex0 3 2) '((a b c  d e f)
                       (g h i  j k l)
                       (m n o  p q r)
                       (s t u  v w x)
                       (y z 0  1 2 3)
                       (4 5 6  7 8 9)))

(test (cols ex0 3 2) '((a g m s y 4)
                       (b h n t z 5)
                       (c i o u 0 6)
                       (d j p v 1 7)
                       (e k q w 2 8)
                       (f l r x 3 9)))


(test (sections ex0 3 2) '((a b c g h i)
                           (d e f j k l)
                           (m n o s t u)
                           (p q r v w x)
                           (y z 0 4 5 6)
                           (1 2 3 7 8 9)))

(test #(1 2 3 4 5) (update-cell #(1 2 3 4 4) 4 5))

(test (ok-set? '(1 2 3 4)) #t)
(test (ok-set? '(1 2 #f 4)) #t)
(test (ok-set? '(1 2 4 4)) #f)

(define ex1 #(1 2 3  4 5 6
              4 5 6  1 2 3
        
              2 3 4  5 6 1
              5 6 1  2 3 4
              
              3 4 5  6 1 2
              6 1 2  3 4 5))

(define ex2 #(1 2 3  4 5 6
              4 5 6  1 2 3
        
              2 3 4  #f #f #f
              5 6 1  #f #f #f
              
              3 4 5  6 1 2
              6 1 2  3 4 5))

(define ex3 #(1 2 3  4 5 6
              4 5 6  1 2 3
              ;;  V
              2 3 5  #f #f #f
              5 6 1  #f #f #f
              
              3 4 #f  6 1 2
              6 1 2  3 4 5))

(define ex4 #(1 2 3  4 5 6
              4 5 6  1 2 3
              ;;  V
              2 3 5  #f #f #f
              #f 6 1  #f #f #f
              
              3 4 5  6 1 2
              6 1 2  3 4 5))

(define ex5 #(1 2 3   4 5 6
              4 5 6   1 2 3
              ;;  V
              2 3 5   5 #f #f
              #f 6 1  #f #f #f
              
              3 4 #f  6 1 2
              6 1 2   3 4 5))


(test (ok? ex2 3 2) #t)
(test (ok? ex3 3 2) #f)
(test (ok? ex4 3 2) #f)
(test (ok? ex5 3 2) #f)

(test (solve ex2 3 2 0) ex1)

(test (solve/k ex2 3 2 0 (lambda (v f) v) (lambda () #f)) ex1)

; ----------------------------------------

;; count-solutions: board dimen dimen -> int
(define (count-solutions board m n)
  (let ([cnt 0])
    (solve/k board m n 0
             (lambda (v f)
               (set! cnt (add1 cnt))
               (f))
             (lambda () cnt))))

(test 288 (count-solutions (make-vector 16 #f) 2 2))
; (count-solutions (make-vector 81 #f) 3 3)

(run-tests)
